# datagrepper - HTTP API for datanommer and the fedmsg bus
# Copyright (C) 2013  Red Hat, Inc. and others
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
import codecs
import itertools
import json
import os
import re
import traceback
from datetime import datetime, timedelta

import arrow
import datanommer.models as dm
import docutils
import docutils.examples
import flask
import jinja2
import markupsafe
import pygal
import pygments
import pygments.formatters
import pygments.lexers
from flask import Flask
from flask_healthz import HealthError, healthz
from pkg_resources import get_distribution
from werkzeug.exceptions import BadRequest

from datagrepper.util import (
    as_bool,
    assemble_timerange,
    DateAwareJSONEncoder,
    get_message_dict,
    json_return,
    message_card,
    request_wants_html,
)


app = Flask(__name__)
app.config.from_object("datagrepper.default_config")
if "DATAGREPPER_CONFIG" in os.environ:
    app.config.from_envvar("DATAGREPPER_CONFIG")
app.config["CORS_DOMAINS"] = list(map(re.compile, app.config.get("CORS_DOMAINS", [])))
app.config["CORS_HEADERS"] = list(map(re.compile, app.config.get("CORS_HEADERS", [])))

# Initialize a datanommer session.
dm.init(app.config.get("DATANOMMER_SQLALCHEMY_URL"))

# Register views
app.register_blueprint(healthz, url_prefix="/healthz")

import datagrepper.widgets  # noqa: E402,F401


@app.context_processor
def inject_variable():
    """Inject some global variables into all templates"""
    extras = {
        "models_version": get_distribution("datanommer.models").version,
        "grepper_version": get_distribution("datagrepper").version,
    }
    return extras


def match_regex_list(val, regexes):
    for regex in regexes:
        if regex.match(val):
            return True
    return False


def filter_regex_list(vals, regexes):
    return [val for val in vals if match_regex_list(val, regexes)]


@app.after_request
def add_cors(response):
    """Allow CORS for domains specified in the config"""
    if "Origin" in flask.request.headers:
        # Handle a CORS request
        origin = flask.request.headers["Origin"]
        if flask.request.method in app.config["CORS_METHODS"] and match_regex_list(
            origin, app.config["CORS_DOMAINS"]
        ):
            response.headers["Access-Control-Allow-Origin"] = origin
            if "Access-Control-Request-Method" in flask.request.headers:
                response.headers["Access-Control-Allow-Methods"] = ", ".join(
                    app.config["CORS_METHODS"]
                )
            if "Access-Control-Request-Headers" in flask.request.headers:
                requested_headers = flask.request.headers[
                    "Access-Control-Request-Headers"
                ]
                requested_headers = [h.strip() for h in requested_headers.split(",")]
                allowed_headers = filter_regex_list(
                    requested_headers, app.config["CORS_HEADERS"]
                )
                if allowed_headers:
                    response.headers["Access-Control-Allow-Headers"] = ", ".join(
                        allowed_headers
                    )
            if flask.request.method == "OPTIONS":
                # Special handling for pre-flight requests
                if "Vary" in response.headers:
                    response.headers["Vary"] = response.headers["Vary"] + ", Origin"
                else:
                    response.headers["Vary"] = "Origin"
                response.headers["Access-Control-Max-Age"] = app.config["CORS_MAX_AGE"]
    return response


@app.teardown_appcontext
def remove_session(exc):
    """Remove the session, which rolls back the transaction in progress.

    This is safe because Datagrepper never makes modifications to the database.
    """
    dm.session.remove()


def modify_rst(rst):
    """Downgrade some of our rst directives if docutils is too old."""

    try:
        # The rst features we need were introduced in this version
        minimum = [0, 9]
        version = map(int, docutils.__version__.split("."))

        # If we're at or later than that version, no need to downgrade
        if version >= minimum:
            return rst
    except Exception:
        # If there was some error parsing or comparing versions, run the
        # substitutions just to be safe.
        pass

    # Otherwise, make code-blocks into just literal blocks.
    substitutions = {
        ".. code-block:: javascript": "::",
    }
    for old, new in substitutions.items():
        rst = rst.replace(old, new)

    return rst


def modify_html(html):
    """Perform style substitutions where docutils doesn't do what we want."""

    substitutions = {
        '<tt class="docutils literal">': "<code>",
        "</tt>": "</code>",
    }
    for old, new in substitutions.items():
        html = html.replace(old, new)

    return html


def preload_docs(endpoint):
    """Utility to load an RST file and turn it into fancy HTML."""

    here = os.path.dirname(os.path.abspath(__file__))
    default = os.path.join(here, "docs")
    directory = app.config.get("DATAGREPPER_DOC_PATH", default)
    fname = os.path.join(directory, endpoint + ".rst")
    with codecs.open(fname, "r", "utf-8") as f:
        rst = f.read()

    rst = modify_rst(rst)

    api_docs = docutils.examples.html_body(rst)

    api_docs = modify_html(api_docs)

    api_docs = markupsafe.Markup(api_docs)
    return api_docs


htmldocs = dict.fromkeys(["index", "reference", "widget", "charts"])
for key in htmldocs:
    htmldocs[key] = preload_docs(key)


def load_docs(request):
    URL = app.config.get("DATAGREPPER_BASE_URL", request.url_root)
    docs = htmldocs[request.endpoint]
    docs = jinja2.Template(docs).render(URL=URL)
    return markupsafe.Markup(docs)


def count_all_messages():
    """Return a count of all messages in the db.

    In some cases, doing a full count of all the message on a postgres database
    takes too long (many tens of seconds).  We can produce a much faster query
    like this, but it only returns the approximate number of messages in the
    db.
    """

    if app.config.get("DATAGREPPER_APPROXIMATE_COUNT"):
        query = "SELECT * FROM approximate_row_count('messages');"
        total = dm.session.execute(query).first()[0]
    else:
        total = dm.Message.grep(defer=True)[0]

    return int(total)


POSSIBLE_SIZES = ["small", "medium", "large", "extra-large"]


@app.route("/")
def index():
    total = count_all_messages()
    docs = load_docs(flask.request)
    return flask.render_template("index.html", docs=docs, total=total)


@app.route("/reference/")
@app.route("/reference")
def reference():
    return flask.render_template("index.html", docs=load_docs(flask.request))


@app.route("/charts/")
@app.route("/charts")
def charts():
    return flask.render_template("index.html", docs=load_docs(flask.request))


@app.route("/widget/")
@app.route("/widget")
def widget():
    return flask.render_template("index.html", docs=load_docs(flask.request))


@app.route("/raw", methods=["GET"], strict_slashes=False)
@app.route("/v2/search", methods=["GET"], strict_slashes=False)
def raw():
    """Main API entry point."""

    # Perform our complicated datetime logic
    start = flask.request.args.get("start", None)
    end = flask.request.args.get("end", None)
    delta = flask.request.args.get("delta")
    start, end, delta = assemble_timerange(start, end, delta)

    # Further filters, all ANDed together in CNF style.
    users = flask.request.args.getlist("user")
    packages = flask.request.args.getlist("package")
    categories = flask.request.args.getlist("category")
    topics = flask.request.args.getlist("topic")
    contains = flask.request.args.getlist("contains")
    # Validate the "contains" argument
    _contains_limit = datetime.utcnow() - timedelta(weeks=4 * 8)
    if contains and datetime.fromtimestamp(start or 0) < _contains_limit:
        raise BadRequest(
            "When using contains, specify a start at most eight months into the past"
        )
    if contains and not (categories or topics):
        raise BadRequest(
            "When using contains, specify either a topic or a category as well"
        )

    # Still more filters.. negations of the previous ones.
    not_users = flask.request.args.getlist("not_user")
    not_packages = flask.request.args.getlist("not_package")
    not_categories = flask.request.args.getlist("not_category")
    not_topics = flask.request.args.getlist("not_topic")

    # Paging arguments
    page = int(flask.request.args.get("page", 1))
    if page < 1:
        raise BadRequest("page must be > 0")
    rows_per_page = int(flask.request.args.get("rows_per_page", 25))
    if rows_per_page > 100:
        raise BadRequest("rows_per_page must be <= 100")
    order = flask.request.args.get("order", "desc")
    if order not in ["desc", "asc"]:
        raise BadRequest("order must be either 'desc' or 'asc'")
    # adding size as paging arguments
    size = flask.request.args.get("size", "large")
    if size not in POSSIBLE_SIZES:
        raise BadRequest(f"size must be in one of these: {POSSIBLE_SIZES!r}")
    # adding chrome as paging arguments
    try:
        chrome = flask.request.args.get("chrome", "true", as_bool)
    except ValueError:
        raise BadRequest("chrome should be either 'true' or 'false'")

    # Response formatting arguments
    callback = flask.request.args.get("callback", None)
    meta = flask.request.args.getlist("meta")

    arguments = dict(
        start=start,
        delta=delta,
        end=end,
        users=users,
        packages=packages,
        categories=categories,
        topics=topics,
        contains=contains,
        not_users=not_users,
        not_packages=not_packages,
        not_categories=not_categories,
        not_topics=not_topics,
        page=page,
        rows_per_page=rows_per_page,
        order=order,
        meta=meta,
    )

    is_html = request_wants_html() and not callback

    try:
        # This fancy classmethod does all of our search for us.
        total, pages, messages = dm.Message.grep(
            start=start and datetime.fromtimestamp(start),
            end=end and datetime.fromtimestamp(end),
            page=page,
            rows_per_page=rows_per_page,
            order=order,
            users=users,
            packages=packages,
            categories=categories,
            topics=topics,
            contains=contains,
            not_users=not_users,
            not_packages=not_packages,
            not_categories=not_categories,
            not_topics=not_topics,
        )
    except Exception as e:
        traceback.print_exc()

        # TODO: return HTML when is_html is True
        output = dict(
            error=str(e),
            arguments=arguments,
        )
        # :D
        if app.config.get("DEBUG", False):
            output["tb"] = traceback.format_exc().split("\n")
        return json_return(output, 500, callback)

    # return HTML content else json
    if is_html:
        # removes boilerlate codes if chrome value is false
        template = "base.html" if chrome else "raw.html"
        return flask.render_template(
            template,
            size=size,
            response=[message_card(msg) for msg in messages],
            arguments=arguments,
            autoscroll=chrome,
        )

    else:
        output = dict(
            raw_messages=[get_message_dict(msg, meta) for msg in messages],
            total=total,
            pages=pages,
            count=len(messages),
            arguments=arguments,
        )
        return json_return(output, 200, callback)


# Get a message by msg_id
@app.route("/id", methods=["GET"], strict_slashes=False)
@app.route("/v2/id", methods=["GET"], strict_slashes=False)
def msg_id():
    if "id" not in flask.request.args:
        flask.abort(400)
    msg = dm.Message.query.filter_by(msg_id=flask.request.args["id"]).first()
    if not msg:
        flask.abort(404)

    # get paging argument for size and chrome
    size = flask.request.args.get("size", "large")
    if size not in POSSIBLE_SIZES:
        raise ValueError(f"size must be in one of these: {POSSIBLE_SIZES!r}")
    try:
        chrome = flask.request.args.get("chrome", "true", as_bool)
    except ValueError:
        raise ValueError("chrome should be either 'true' or 'false'")
    # get paging argument for is_raw
    # is_raw checks if card comes from the search endpoint
    try:
        is_raw = flask.request.args.get("is_raw", "false", as_bool)
    except ValueError:
        raise ValueError("is_raw should be either 'true' or 'false'")

    callback = flask.request.args.get("callback", None)
    meta = flask.request.args.getlist("meta")

    # converts message from sqlalchemy objects to json-like dicts
    msg_dict = get_message_dict(msg, meta)

    if request_wants_html() and not callback:
        # convert string into python dictionary
        msg_string = pygments.highlight(
            json.dumps(msg_dict, indent=2, sort_keys=True, cls=DateAwareJSONEncoder),
            pygments.lexers.JavascriptLexer(),
            pygments.formatters.HtmlFormatter(
                noclasses=True,
                style="emacs",
            ),
        ).strip()

        template = "base.html" if chrome else "raw.html"
        return flask.render_template(
            template,
            size=size,
            response=[message_card(msg)],
            msg_string=msg_string,
            heading="Message by ID",
            is_raw=is_raw,
        )
    else:
        return json_return(msg_dict, callback=callback)


@app.route("/charts/<chart_type>", methods=["GET"])
def make_charts(chart_type):
    """Return SVGs graphing db content."""

    # Perform our complicated datetime logic
    start = flask.request.args.get("start", None)
    end = flask.request.args.get("end", None)
    delta = flask.request.args.get("delta", None)
    start, end, delta = assemble_timerange(start, end, delta)

    # Further filters, all ANDed together in CNF style.
    users = flask.request.args.getlist("user")
    packages = flask.request.args.getlist("package")
    categories = flask.request.args.getlist("category")
    topics = flask.request.args.getlist("topic")
    contains = flask.request.args.getlist("contains")

    # Still more filters.. negations of the previous ones.
    not_users = flask.request.args.getlist("not_user")
    not_packages = flask.request.args.getlist("not_package")
    not_categories = flask.request.args.getlist("not_category")
    not_topics = flask.request.args.getlist("not_topic")

    end = end and datetime.fromtimestamp(end)
    start = start and datetime.fromtimestamp(start)
    end = end or datetime.utcnow()
    start = start or end - timedelta(days=365)

    human_readable = flask.request.args.get("human_readable", True, as_bool)
    logarithmic = flask.request.args.get("logarithmic", False, as_bool)
    show_x_labels = flask.request.args.get("show_x_labels", True, as_bool)
    show_y_labels = flask.request.args.get("show_y_labels", True, as_bool)
    show_dots = flask.request.args.get("show_dots", True, as_bool)
    fill = flask.request.args.get("fill", False, as_bool)

    title = flask.request.args.get("title", "fedmsg events")
    width = flask.request.args.get("width", 800, int)
    height = flask.request.args.get("height", 800, int)

    interpolation = flask.request.args.get("interpolation", None)
    interpolation_types = [
        None,
        "quadratic",
        "cubic",
    ]
    if interpolation not in interpolation_types:
        flask.abort(404, f"{interpolation} not in {interpolation_types!r}")

    chart_types = {
        "line": "Line",
        "stackedline": "StackedLine",
        "xy": "XY",
        "bar": "Bar",
        "horizontalbar": "HorizontalBar",
        "stackedbar": "StackedBar",
        "horizontalstackedbar": "HorizontalStackedBar",
        "funnel": "Funnel",
        "pyramid": "Pyramid",
        "verticalpyramid": "VerticalPyramid",
        "dot": "Dot",
        "gauge": "Gauge",
    }
    if chart_type not in chart_types:
        flask.abort(404, f"{chart_type} not in {chart_types!r}")

    style = flask.request.args.get("style", "default")
    if style not in pygal.style.styles:
        flask.abort(404, f"{style} not in {pygal.style.styles!r}")
    style = pygal.style.styles[style]

    chart = getattr(pygal, chart_types[chart_type])(
        human_readable=human_readable,
        logarithmic=logarithmic,
        show_x_labels=show_x_labels,
        show_y_labels=show_y_labels,
        show_dots=show_dots,
        fill=fill,
        title=title,
        width=width,
        height=height,
        interpolate=interpolation,
        x_label_rotation=45,
        style=style,
    )

    lookup = locals()
    factor_names = flask.request.args.getlist("split_on")
    factor_names = [name for name in factor_names if lookup[name]]
    factor_values = [lookup[name] for name in factor_names]
    factors = list(itertools.product(*factor_values))

    N = int(flask.request.args.get("N", 10))
    if N < 3:
        flask.abort(500, "N must be greater than 3")
    if N > 15:
        flask.abort(500, "N must be less than 15")

    try:
        labels = []

        kwargs = dict(
            users=users,
            packages=packages,
            categories=categories,
            topics=topics,
            contains=contains,
        )

        dates = [i for i, _ in daterange(start, end, N)]
        if human_readable:
            labels = [arrow.get(i).humanize() for i in dates]
        else:
            labels = [str(arrow.get(i).date()) for i in dates]

        for factor in factors:
            for i, name in enumerate(factor_names):
                kwargs[name] = [factor[i]]

            values = []

            for i, j in daterange(start, end, N):
                count, _, _ = dm.Message.grep(
                    start=i,
                    end=j,
                    rows_per_page=None,
                    defer=True,
                    not_users=not_users,
                    not_packages=not_packages,
                    not_categories=not_categories,
                    not_topics=not_topics,
                    **kwargs,
                )
                values.append(count)
            tag = factor and " & ".join(factor) or "events"

            # Truncate things to make charts prettier
            if tag.startswith("org.fedoraproject.prod."):
                tag = tag[len("org.fedoraproject.prod.") :]

            chart.add(tag, values)

        chart.x_labels = labels
        output = chart.render()
        status = 200
        mimetype = "image/svg+xml"
    except Exception as e:
        import traceback

        traceback.print_exc()
        output = "Error, %r" % e
        status = 500
        mimetype = "text/html"

    return flask.Response(
        response=output,
        status=status,
        mimetype=mimetype,
    )


def daterange(start, stop, steps):
    """A generator for stepping through time."""
    delta = (stop - start) / steps
    current = start
    while current + delta <= stop:
        yield current, current + delta
        current += delta


@app.route("/messagecount", methods=["POST"])
def post_messagecount():
    flask.abort(405)


# Instant requests
@app.route("/messagecount/")
@app.route("/messagecount")
def messagecount():
    total = {}
    total["messagecount"] = count_all_messages()
    total = flask.jsonify(total)
    return total


@app.errorhandler(404)
def not_found(error):  # TODO test this
    return flask.Response(
        response=json.dumps({"error": "not_found"}),
        status=404,
        mimetype="application/json",
    )


@app.errorhandler(500)
def internal_error(error):  # TODO test this
    return flask.Response(
        response=json.dumps(
            {
                "error": "internal_error",
                "detail": str(error),
                "traceback": traceback.format_exc(),
            }
        ),
        status=500,
        mimetype="application/json",
    )


def liveness():
    pass


def readiness():
    try:
        dm.session.execute("SELECT 1")
    except Exception:
        raise HealthError("Can't connect to the database")



APP_PATH = "https://apps.fedoraproject.org/datagrepper"
CORS_DOMAINS = [".*"]
CORS_METHODS = ["GET", "OPTIONS"]
CORS_HEADERS = [".*"]
CORS_MAX_AGE = "600"
DEFAULT_QUERY_DELTA = 0
DATANOMMER_SQLALCHEMY_URL = "postgresql://datanommer:datanommer@localhost/messages"
DATAGREPPER_APPROXIMATE_COUNT = True
BUS_INFO = {
    "link": "https://fedora-messaging.readthedocs.io",
    "shortname": "Fedora Messaging",
    "longname": "Fedora Messaging bus",
    "docs": "https://fedora-messaging.readthedocs.io",
}
THEME_CSS_URL = (
    "https://apps.fedoraproject.org/global/fedora-bootstrap-1.0/"
    "fedora-bootstrap.min.css"
)

# Example: "wss://hub.fedoraproject.org:9939"
WEBSOCKET_URL = None
# Only allow websockets connections to fedoraproject.org, for instance:
# "connect-src https://*.fedoraproject.org wss://*.fedoraproject.org"
CONTENT_SECURITY_POLICY = None

HEALTHZ = {
    "live": "datagrepper.app.liveness",
    "ready": "datagrepper.app.readiness",
}



# SPDX-License-Identifier: GPL-2.0+
# Copyright 2018 Mike Bonnet <mikeb@redhat.com>

import json
import unittest
from datetime import datetime, timedelta, timezone
from unittest.mock import MagicMock, patch

import datagrepper.app


class TestAPI(unittest.TestCase):
    def setUp(self):
        datagrepper.app.app.testing = True
        self.client = datagrepper.app.app.test_client()

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_reference(self, grep):
        resp = self.client.get("/reference")
        self.assertEqual(resp.status_code, 200)
        response = resp.get_data()
        self.assertIn(b"General API notes", response)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_charts(self, grep):
        resp = self.client.get("/charts")
        self.assertEqual(resp.status_code, 200)
        response = resp.get_data()
        self.assertIn(b"Charts and Graphs", response)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_widget(self, grep):
        resp = self.client.get("/widget")
        self.assertEqual(resp.status_code, 200)
        response = resp.get_data()
        self.assertIn(b"Embeddable Widget", response)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_defaults(self, grep):
        resp = self.client.get("/raw")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(grep.call_args[0], ())
        kws = grep.call_args[1]
        self.assertIsNone(kws["start"])
        self.assertIsNone(kws["end"])
        self.assertEqual(kws["page"], 1)
        self.assertEqual(kws["rows_per_page"], 25)
        self.assertEqual(kws["order"], "desc")
        for arg in [
            "users",
            "packages",
            "categories",
            "topics",
            "contains",
            "not_users",
            "not_packages",
            "not_categories",
            "not_topics",
        ]:
            self.assertEqual(kws[arg], [])

    @patch("datagrepper.app.count_all_messages", autospec=True, return_value=42)
    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_index(self, grep, count_all_messages):
        resp = self.client.get("/")
        self.assertEqual(resp.status_code, 200)
        response = resp.get_data()
        self.assertIn(b"Datagrepper's documentation", response)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_default_result(self, grep):
        resp = self.client.get("/raw")
        self.assertEqual(resp.status_code, 200)
        result = json.loads(resp.get_data())
        self.assertEqual(result["count"], 0)
        self.assertEqual(result["pages"], 0)
        self.assertEqual(result["total"], 0)
        self.assertEqual(result["raw_messages"], [])

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_contains_delta(self, grep):
        # At one point, this would produce a traceback/500.
        resp = self.client.get("/raw?delta=14400&category=wat&contains=foo")
        self.assertEqual(resp.status_code, 200)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_contains_delta_and_start(self, grep):
        resp = self.client.get("/raw?start=1564503781&delta=600")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(grep.call_args[0], ())
        kws = grep.call_args[1]

        expected_start = datetime.fromtimestamp(1564503781)
        self.assertEqual(kws["start"], expected_start)

        expected_end = expected_start + timedelta(seconds=600)
        self.assertEqual(kws["end"], expected_end)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 180})
    def test_raw_default_query_delta(self, grep):
        resp = self.client.get("/raw")
        self.assertEqual(resp.status_code, 200)
        kws = grep.call_args[1]
        # Verify the default query delta was applied
        self.assertEqual((kws["end"] - kws["start"]).total_seconds(), 180.0)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 180})
    def test_raw_default_query_delta_with_start(self, grep):
        resp = self.client.get("/raw?start=1564503781")
        self.assertEqual(resp.status_code, 200)
        kws = grep.call_args[1]
        # Verify the default query delta was not applied
        self.assertNotEqual((kws["end"] - kws["start"]).total_seconds(), 180.0)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 180})
    def test_raw_default_query_delta_with_delta(self, grep):
        resp = self.client.get("/raw?delta=7200")
        self.assertEqual(resp.status_code, 200)
        kws = grep.call_args[1]
        # Verify the default query delta was not applied
        self.assertEqual((kws["end"] - kws["start"]).total_seconds(), 7200.0)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 180})
    def test_raw_default_query_delta_with_end(self, grep):
        resp = self.client.get("/raw?end=1564503781")
        self.assertEqual(resp.status_code, 200)
        kws = grep.call_args[1]
        self.assertEqual((kws["end"] - kws["start"]).total_seconds(), 180.0)

    @patch("datagrepper.app.dm.Message.grep", side_effect=Exception)
    def test_raw_exception(self, grep):
        resp = self.client.get("/raw")
        self.assertEqual(resp.status_code, 500)
        response = resp.get_data()
        self.assertNotIn(b'"tb": ["Traceback', response)

    @patch("datagrepper.app.dm.Message.grep", side_effect=Exception)
    @patch.dict(datagrepper.app.app.config, {"DEBUG": True})
    def test_raw_exception_debug_true(self, grep):
        resp = self.client.get("/raw")
        self.assertEqual(resp.status_code, 500)
        response = resp.get_data()
        self.assertIn(b'"tb": ["Traceback', response)
        self.assertIn(b"Exception", response)

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_default_query_with_start_and_end_native(self, grep):
        resp = self.client.get(
            "/raw?start=2012-01-02T03:04:05%2B00:00&end=2012-01-02T04:04:05%2B00:00"
        )
        self.assertEqual(resp.status_code, 200)
        kws = grep.call_args[1]
        self.assertEqual(
            kws["start"].astimezone(timezone.utc),
            datetime(2012, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        )
        self.assertEqual(
            kws["end"].astimezone(timezone.utc),
            datetime(2012, 1, 2, 4, 4, 5, tzinfo=timezone.utc),
        )

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_contains_without_delta(self, grep):
        """https://github.com/fedora-infra/datagrepper/issues/206"""
        resp = self.client.get("/raw?category=wat&contains=foo")
        self.assertEqual(resp.status_code, 400)
        target = b"When using contains, specify a start at most eight months"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_exceptions_contains_and_not_categories_topics(self, grep):
        timestamp = datetime.utcnow()
        resp = self.client.get(f"/raw?contains=hello&start={timestamp}")
        self.assertEqual(resp.status_code, 400)
        target = b"When using contains, specify either a topic or a category as well"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_exceptions_page_less_than_zero(self, grep):
        resp = self.client.get("/raw?page=0")
        self.assertEqual(resp.status_code, 400)
        target = b"page must be &gt; 0"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_exceptions_rows_per_page_over_100(self, grep):
        resp = self.client.get("/raw?rows_per_page=101")
        self.assertEqual(resp.status_code, 400)
        target = b"rows_per_page must be &lt;= 100"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_exceptions_order_not_in_list(self, grep):
        resp = self.client.get("/raw?order=notinlist")
        self.assertEqual(resp.status_code, 400)
        target = b"order must be either &#x27;desc&#x27; or &#x27;asc&#x27;"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_raw_exceptions_possible_sizes_not_in_list(self, grep):
        resp = self.client.get("/raw?size=suchastring")
        self.assertEqual(resp.status_code, 400)
        target = b"size must be in one of these:"
        assert target in resp.data, f"{target!r} not in {resp.data!r}"

    @patch("datagrepper.app.dm.Message.query", autospec=True)
    def test_id(self, query):
        msg = query.filter_by.return_value.first.return_value
        msg.as_dict = MagicMock(return_value={"key": "value"})
        resp = self.client.get("/id?id=one")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(query.filter_by.call_args, ((), {"msg_id": "one"}))

    @patch("datagrepper.app.count_all_messages", autospec=True, return_value=42)
    def test_count(self, count_all_messages):
        resp = self.client.get("/messagecount")
        self.assertEqual(resp.status_code, 200)
        result = json.loads(resp.get_data())
        self.assertEqual(result, {"messagecount": 42})

    @patch("datagrepper.app.dm.Message.grep", return_value=(0, 0, []))
    def test_chart_line(self, grep):
        resp = self.client.get("/charts/line")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.mimetype, "image/svg+xml")
        svg_data = (
            b'<svg xmlns="http://www.w3.org/2000/svg" '
            b'xmlns:xlink="http://www.w3.org/1999/xlink"'
        )
        # This bloc is to get rid of the downstream patch.
        # The output is different depending on lxml.
        try:
            import lxml  # noqa
        except ImportError:
            svg_data = b'<svg xmlns:xlink="http://www.w3.org/1999/xlink"'
        self.assertIn(
            svg_data,
            resp.get_data(),
        )

    def test_healthz_liveness(self):
        """Test the /healthz/live check endpoint"""
        resp = self.client.get("/healthz/live")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, b'{"status": 200, "title": "OK"}')

    @patch("datagrepper.app.dm.session.execute")
    def test_healthz_readiness_ok(self, execute):
        """Test the /healthz/ready check endpoint"""
        resp = self.client.get("/healthz/ready")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, b'{"status": 200, "title": "OK"}')

    @patch("datagrepper.app.dm.session.execute", side_effect=Exception)
    def test_healthz_readiness_not_ok(self, execute):
        """Test the /healthz/ready check endpoint when not ready"""
        resp = self.client.get("/healthz/ready")
        self.assertEqual(resp.status_code, 503)
        self.assertEqual(
            resp.data, b'{"status": 503, "title": "Can\'t connect to the database"}'
        )

    def test_widget_js(self):
        resp = self.client.get("/widget.js")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.mimetype, "application/javascript")
        response = resp.get_data()
        self.assertIn(b"$('script#datagrepper-widget')", response)
        self.assertIn(b"datagrepper-widget", response)

    def test_widget_css(self):
        resp = self.client.get("/widget.js?css=true")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.mimetype, "application/javascript")
        response = resp.get_data()
        self.assertIn(b"/static/css/bootstrap.css", response)



""" Contains code for producing embeddable widgets """

import flask

from datagrepper.app import app


js_helpers = """
function include_js(url, success) {
    var script     = document.createElement('script');
    script.src = url;

    var head = document.getElementsByTagName('head')[0],
    done = false;
    // Attach handlers for all browsers
    script.onload = script.onreadystatechange = function() {
        if (!done && (
                !this.readyState ||
                this.readyState == 'loaded' ||
                this.readyState == 'complete'
                )) {
            done = true;
            success();  // Do the callback
            script.onload = script.onreadystatechange = null;
            head.removeChild(script);  /* poof */
        };
    };
    head.appendChild(script);
};

function run_with_jquery(callback) {
    var jq_url = '%(base)s/static/jquery-2.1.0.min.js'
    if (typeof jQuery == 'undefined') {
        include_js(jq_url, callback);
    } else {
        callback();
    }
}"""

work = """
var datagrepper_success = function(json) {
    $.each(json.raw_messages, function(i, msg) {
        var meta = msg.meta;
        if (meta === undefined) { meta = msg; }
        var card = '<div class="message-card">';
        if (meta.icon) {
            if (meta.link) {
                card = card +
                    '<a href="' + meta.link + '">' +
                    '<img src="' + meta.icon + '"/>' +
                    '</a>';
            } else {
                card = card + '<img src="' + meta.icon + '"/>';
            }
        }
        if (meta.secondary_icon) {
            card = card +
                '<img src="' + meta.secondary_icon + '"/>';
        }
        card = card +
            '<p>' + meta.subtitle + '</p>' +
            '</div>';
        card = card + '<div class="datetime">' + meta['date'] + '</div>';
        $("div#datagrepper-widget").append(card);
    });
}

// These are the default arguments that we use for our datagrepper query
var data = {
    order: 'desc',
    chrome: 'false'
};

// Check to see if the user has asked us to filter the firehose.
var datagrepper_attrs = [
    'user', 'package', 'category', 'topic',
    'order', 'rows_per_page', 'page', 'size'];
$.each(datagrepper_attrs, function(i, attr) {
    var value = $('script#datagrepper-widget').attr("data-" + attr);
    if (value != undefined) {
        data[attr] = value;
    }
});

$.ajax(
    '%(base)s/v2/search?meta=url&meta=app_icon' +
    '&meta=agent_avatar&meta=summary&meta=date', {
        data: data,
        dataType: 'jsonp',
        success: datagrepper_success,
        error: function() {
            console.log(arguments);
        }
    }
)
"""

css_helper = """
$('head').append('<link rel="stylesheet" href="%s" type="text/css"/>');
"""


@app.route("/widget.js")
def widget_js():
    """This code is super ugly.
    But it produces a widget as a self-extracting script.

    """

    prefix = flask.request.url_root[:-1]

    raw_widget = '<div id="datagrepper-widget"></div>'
    scripts, calls, css = [], [work % dict(base=prefix)], []

    if flask.request.args.get("css", "").lower() == "true":

        def static_url(filename):
            return app.config["APP_PATH"] + "/static/" + filename

        css.append(css_helper % static_url("css/bootstrap.css"))
        css.append(css_helper % static_url("css/raw.css"))

    # This, ridiculously, will find the place in the DOM of the script tag
    # responsible for running this javascript at the time of its execution.
    # This allows us to inject our graph in-place; i.e., wherever the user
    # includes our tag, that's where the graph will unpack itself.
    calls.insert(0, "$('script#datagrepper-widget').before('%s');" % raw_widget.strip())
    calls.extend(css)

    # Just for debugging...
    # calls.append("console.log('waaaaaat!');")
    inner_payload = ";\n".join(calls)

    envelope = inner_payload
    for script in reversed(scripts):
        envelope = """$.getScript("{}", function(){{{}}});""".format(
            prefix + script,
            envelope,
        )  # TODO test this

    body = js_helpers % dict(base=prefix)
    body += "\nrun_with_jquery(function() {%s});" % envelope

    return flask.Response(
        response=body,
        status=200,
        mimetype="application/javascript",
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datagrepper', 'datagrepper.docs', 'tests']

package_data = \
{'': ['*'],
 'datagrepper': ['static/*',
                 'static/css/*',
                 'static/css/fonts/Bold/*',
                 'static/css/fonts/BoldItalic/*',
                 'static/css/fonts/Italic/*',
                 'static/css/fonts/Light/*',
                 'static/css/fonts/LightItalic/*',
                 'static/css/fonts/Regular/*',
                 'templates/*']}

install_requires = \
['Pygments>=2.9.0,<3.0.0',
 'SQLAlchemy>=1.4.0,<2.0.0',
 'arrow>=1.1.1,<2.0.0',
 'cffi>=1.14.6,<2.0.0',
 'datanommer.models>=1.0.0,<2.0.0',
 'docutils>=0.16',
 'fedora-messaging>=2.1.0',
 'flask-healthz>=0.0.3,<0.0.4',
 'flask>=2.0.1',
 'moksha.common>=1.2.5,<2.0.0',
 'psycopg2>=2.9.1,<3.0.0',
 'pygal>=2.4.0',
 'python-dateutil>=2.8.1,<3.0.0']

extras_require = \
{'deploy': ['gunicorn>=20.0,<21.0'],
 'schemas': ['anitya-schema',
             'bodhi-messages',
             'copr-messaging',
             'discourse2fedmsg-messages',
             'fedocal-messages',
             'fedorainfra-ansible-messages',
             'fedora-messaging-the-new-hotness-schema',
             'fedora-planet-messages',
             'mdapi-messages',
             'noggin-messages',
             'nuancier-messages',
             'pagure-messages']}

setup_kwargs = {
    'name': 'datagrepper',
    'version': '1.0.1',
    'description': 'A webapp to query datanommer',
    'long_description': "# datagrepper\n\nDatagrepper is a web application and JSON API to retrieve historical messages sent via Fedora Messaging. [Datanommer](https://github.com/fedora-infra/datanommer/) is a seperate project and service that consumes messages from the Fedora Messaging queue and puts them in a database. These messages is what datagrepper queries. \n\nDatagrepper is curently running in production at https://apps.fedoraproject.org/datagrepper/\n\n## Development Environment\n\nVagrant allows contributors to get quickly up and running with a datagrepper development environment by automatically configuring a virtual machine. \n\nThe datagrepper Vagrant environment configures configures and enables a datanommer service and database. The datanommer instance is configured to be empty when first provisioned, but to consume messages from the stage Fedora Messaging queue.\n\n### Install vagrant\nTo get started, run the following commands to install the Vagrant and Virtualization packages needed, and start the libvirt service:\n\n    $ sudo dnf install ansible libvirt vagrant-libvirt vagrant-sshfs vagrant-hostmanager\n    $ sudo systemctl enable libvirtd\n    $ sudo systemctl start libvirtd\n\n### Checkout and Provision\nNext, check out the datagrepper code and run vagrant up:\n\n    $ git clone https://github.com/fedora-infra/datagrepper\n    $ cd datanommer\n    $ vagrant up\n\n### Interacting with your development datagrepper\nAfter successful provisioning of the Datagrepper vagrant setup, the datagrepper web application will be accessible from your host machine's web browser at\n\nhttp://datagrepper.test:5000/\n\n\n\n\n### Using the development environment\nSSH into your newly provisioned development environment:\n\n    $ vagrant ssh\n\nThe vagrant setup also defines 4 handy commands to interact with the service that runs the datagrepper flask application: \n\n    $ datagrepper-start\n    $ datagrepper-stop\n    $ datagrepper-restart\n    $ dataprepper-logs\n\nAdditionally, the following commands are also available for interacting with the datanommer service:\n\n    $ datanommer-consumer-start\n    $ datanommer-consumer-stop\n    $ datanommer-consumer-restart\n    $ datanommer-consumer-logs\n\n### Running the tests\nDatanommer is comprised of 3 seperate modules in this single repository. There is a handy script in the top directory of this repo to run the tests on all 3 modules:\n\n    $ ./runtests.sh\n\nHowever, tests can also be run on a single module by invotking tox in that modules' directory. For example:\n\n    $ cd datanommer.models/\n    $ tox\n\nNote, that the tests use virtual environments that are not created from scratch with every subsequent run of the tests. Therefore, **when changes happen to dependencies, the tests may fail to run correctly**. To recreate the virtual envrionments,  run the tests commands with the `-r` flag, for example:\n\n    $ ./runtests.sh -r\n\nor\n\n    $ cd datanommer.models/\n    $ tox -r\n",
    'author': 'Fedora Infrastructure',
    'author_email': 'admin@fedoraproject.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fedora-infra/datagrepper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import datetime
import unittest
from unittest.mock import patch

from dateutil import tz

import datagrepper.app
from datagrepper.util import assemble_timerange, datetime_to_seconds


utc = datetime.timezone.utc


class TestTimerange(unittest.TestCase):
    ctx = datagrepper.app.app.app_context()

    def setUp(self):
        self.now = datetime.datetime.fromtimestamp(1325376000, tz=utc)
        patcher = unittest.mock.patch("datagrepper.util.datetime.datetime")
        self.addCleanup(patcher.stop)
        mock_dt = patcher.start()
        # https://docs.python.org/3/library/unittest.mock-examples.html#mock-patching-methods
        mock_dt.now.return_value = self.now
        mock_dt.fromtimestamp.side_effect = (
            lambda *args, **kw: datetime.datetime.fromtimestamp(*args, **kw)
        )
        datagrepper.app.app.testing = True
        self.ctx.push()

    def tearDown(self):
        self.ctx.pop()

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_none_none_none(self):
        start, end, delta = assemble_timerange(None, None, None)
        assert start is None
        assert end is None
        assert delta is None

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 400})
    def test_none_none_none_with_default_delta(self):
        start, end, delta = assemble_timerange(None, None, None)
        assert 1325375600.0 == start
        assert 1325376000.0 == end
        assert 400 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_delta_none_none(self):
        start, end, delta = assemble_timerange(None, None, 5)
        assert 1325375995.0 == start
        assert 1325376000.0 == end
        assert 5 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_none_start_none(self):
        start = self.now - datetime.timedelta(seconds=700)
        start = start.timestamp()
        start, end, delta = assemble_timerange(start, None, None)
        assert 1325375300.0 == start
        assert 1325376000.0 == end
        assert 700 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_delta_start_none(self):
        start = self.now - datetime.timedelta(seconds=600)
        start = start.timestamp()
        start, end, delta = assemble_timerange(start, None, 5)
        assert 1325375400.0 == start
        assert 1325375405.0 == end
        assert 5 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 600})
    def test_none_none_end(self):
        end = self.now - datetime.timedelta(seconds=600)
        end = end.timestamp()
        start, end, delta = assemble_timerange(None, end, None)
        assert 1325374800.0 == start
        assert 1325375400.0 == end
        assert 600 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_delta_none_end(self):
        end = self.now - datetime.timedelta(seconds=600)
        end = end.timestamp()
        start, end, delta = assemble_timerange(None, end, 5)
        assert 1325375395.0 == start
        assert 1325375400.0 == end
        assert 5 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_none_start_end(self):
        end = self.now - datetime.timedelta(seconds=600)
        end = end.timestamp()
        start = self.now - datetime.timedelta(seconds=800)
        start = start.timestamp()
        start, end, delta = assemble_timerange(start, end, None)
        assert 1325375200.0 == start
        assert 1325375400.0 == end
        assert 200 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_delta_start_end(self):
        end = self.now - datetime.timedelta(seconds=600)
        end = end.timestamp()
        start = self.now - datetime.timedelta(seconds=800)
        start = start.timestamp()
        start, end, delta = assemble_timerange(start, end, 5)
        assert 1325375200.0 == start
        assert 1325375400.0 == end
        assert 200 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_none_start_end_parse(self):
        end = "2012-01-01T01:00:00+00:00"
        start = "2012-01-01T00:00:00+00:00"
        start, end, delta = assemble_timerange(start, end, None)
        assert 1325376000.0 == start
        assert 1325376000.0 + 3600 == end
        assert 3600 == delta

    @patch.dict(datagrepper.app.app.config, {"DEFAULT_QUERY_DELTA": 0})
    def test_timedelta_to_seconds(self):
        start = datetime.datetime.now(tz.tzutc())
        time = datetime_to_seconds(start)
        assert time == 1325376000.0



import datetime
import json
import time

import arrow
import flask
from dateutil import tz
from dateutil.parser import parse
from fedora_messaging.message import get_class as get_fm_class


# http://flask.pocoo.org/snippets/45/
# accept header returns json type content only
# However, if the accept header is */*, then return json.
def request_wants_html():
    best = flask.request.accept_mimetypes.best_match(
        ["application/json", "text/html", "text/plain"]
    )
    return best == "text/html"


def json_return(data, status=200, callback=None):
    try:
        output = json.dumps(data, cls=DateAwareJSONEncoder)
    except TypeError:
        flask.current_app.logger.exception(f"Could not encode to JSON: {data!r}")
        raise
    mimetype = flask.request.headers.get("Accept")
    # Our default - http://da.gd/vIIV
    if mimetype == "*/*":
        mimetype = "application/json"

    if callback:
        mimetype = "application/javascript"
        output = f"{callback}({output});"

    return flask.Response(
        response=output,
        status=status,
        mimetype=mimetype,
    )


def datetime_to_seconds(dt):
    """Name this, just because its confusing."""
    return dt.timestamp()


def datetime_to_timestamp(datetime_str_or_timestamp):
    try:
        return float(datetime_str_or_timestamp)
    except ValueError:
        pass

    time = parse(datetime_str_or_timestamp)
    # Assume default timezone is UTC
    if time.tzinfo is None:
        time = time.replace(tzinfo=tz.tzutc())
    return time.timestamp()


def now_seconds():
    return datetime_to_seconds(datetime.datetime.now(tz.tzutc()))


def assemble_timerange(start, end, delta):
    """Util to handle our complicated datetime logic."""

    # Normalize arguments
    if start is not None:
        start = datetime_to_timestamp(start)
    if end is not None:
        end = datetime_to_timestamp(end)
    if delta is not None:
        delta = float(delta)

    default_delta = float(flask.current_app.config["DEFAULT_QUERY_DELTA"])

    # Figure out values for unset arguments.
    valid_args = (start is not None, end is not None, delta is not None)
    if valid_args == (False, False, False):
        if default_delta >= 1:
            end = now_seconds()
            start = end - default_delta
            delta = default_delta
    elif valid_args == (False, False, True):
        end = now_seconds()
        start = end - delta
    elif valid_args == (False, True, False):
        if default_delta >= 1:
            start = end - default_delta
            delta = default_delta
        else:
            start = 0
    elif valid_args == (False, True, True):
        start = end - delta
    elif valid_args == (True, False, False):
        end = now_seconds()
        delta = end - start
    elif valid_args == (True, False, True):
        end = start + delta
    elif valid_args == (True, True, False):
        delta = end - start
    elif valid_args == (True, True, True):
        # Override delta
        delta = end - start

    return start, end, delta


class DateAwareJSONEncoder(json.encoder.JSONEncoder):
    """Encoder with support for datetime objects"""

    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date)):
            # Convert to a UNIX timestamp. I would prefer isoformat but let's not
            # break compatibility.
            return time.mktime(obj.timetuple())
        return super().default(obj)


def as_bool(value):
    if isinstance(value, str):
        value = value.strip().lower()
        if value in ["true", "yes", "on", "y", "t", "1"]:
            return True
        elif value in ["false", "no", "off", "n", "f", "0"]:
            return False
        else:
            raise ValueError(f"value is not true or false: {value}")
    return bool(value)


def get_message_dict(msg, meta):
    # we can drop this if statement once we remove fedmsg
    if flask.request.url_rule.rule.startswith("/v2/"):
        msg_dict = msg.as_fedora_message_dict()
    else:
        msg_dict = msg.as_dict()
    if meta:
        msg_dict["meta"] = meta_argument(msg, meta)
    return msg_dict


def get_fm_message(message):
    """Build a ``fedora_messaging.message.Message`` instance from the DB message instance"""
    headers = message.headers
    if not headers:
        headers = {}
    if "sent-at" not in headers:
        headers["sent-at"] = message.timestamp.isoformat()

    MessageClass = get_fm_class(headers.get("fedora_messaging_schema"))
    fm_message = MessageClass(
        body=message.msg,
        topic=message.topic,
        headers=headers,
        severity=headers.get("fedora_messaging_severity"),
    )
    fm_message.id = message.msg_id
    return fm_message


def message_card(msg):
    """Generate a dict with the message's display information"""
    card = meta_argument(msg, ("date", "url", "summary", "app_icon", "agent_avatar"))
    card["timestamp"] = arrow.get(msg.timestamp)
    # import some keys unchanged
    for key in ("topic", "msg_id"):
        card[key] = getattr(msg, key)
    return card


def meta_argument(msg, meta):
    """Return meta argument values for search and id endpoints
    so that JSON include human-readable strings"""

    meta_expected = {
        "summary",
        "text",
        "url",
        "app_icon",
        "agent_avatar",
        "usernames",
        "packages",
        "containers",
        "modules",
        "flatpaks",
        "date",
    }
    meta_legacy = {
        "subtitle": "summary",
        "link": "url",
        "icon": "app_icon",
    }
    meta_allowed = meta_expected | set(meta_legacy.keys())
    if len(set(meta).intersection(meta_allowed)) != len(set(meta)):
        raise ValueError(
            "meta must be in {}. Got {}".format(
                ",".join(list(meta_allowed)), list(meta)
            )
        )

    fm_msg = get_fm_message(msg)

    metas = {}
    for metadata in meta:
        # This one is exceptional
        if metadata == "date":
            metas[metadata] = arrow.get(msg.timestamp).humanize()
            continue
        # This one is exceptional too ;-)
        if metadata == "text":
            metas[metadata] = str(fm_msg)
            continue
        # Handle legacy (fedmsg) meta
        if metadata in meta_legacy:
            metas[metadata] = getattr(fm_msg, meta_legacy[metadata])
            metas.setdefault("WARNING", []).append(
                f"Meta {metadata} is deprecated and has been replaced by {meta_legacy[metadata]}"
            )
            continue
        # All the other metas use the schema properties
        try:
            metas[metadata] = getattr(fm_msg, metadata)
        except Exception:
            flask.current_app.logger.exception(
                f"Could not get metadata {metadata} for message {msg.id}"
            )
            continue

        # We have to do this because 'set' is not JSON-serializable
        if isinstance(metas[metadata], set):
            metas[metadata] = list(metas[metadata])

    return metas
